/*
 * Copyright (c) 2013, Pavel Lechev
 *    All rights reserved.
 *
 *    Redistribution and use in source and binary forms, with or without modification,
 *    are permitted provided that the following conditions are met:
 *
 *     1) Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *     2) Redistributions in binary form must reproduce the above copyright notice,
 *        this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *     3) Neither the name of the Pavel Lechev nor the names of its contributors may be used to endorse or promote
 *        products derived from this software without specific prior written permission.
 *
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *    IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jmockring.webserver;

import java.io.IOException;
import java.net.URL;
import java.util.List;
import java.util.Properties;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.PostMethod;
import org.eclipse.jetty.server.Connector;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.HandlerCollection;
import org.eclipse.jetty.server.handler.HandlerList;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.util.thread.QueuedThreadPool;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public abstract class AbstractJettyWebServer implements WebServer {

    private static final Logger log = LoggerFactory.getLogger(AbstractJettyWebServer.class);

    private final int port;

    private final int workerThreadCount;

    private final String secret;

    private final String logDir;

    /**
     * @param properties
     * @param port
     */
    public AbstractJettyWebServer(Properties properties, int port) {
        URL location = getClass().getClassLoader().getResource(".");
        this.port = port;
        this.secret = properties.getProperty("jetty.secret", "6ee03119e0b4b0db921b609c18d4d4f1");
        this.logDir = properties.getProperty("log.dir", location.getPath());
        this.workerThreadCount = Integer.valueOf(properties.getProperty("jetty.threads.max", "250"));
    }

    protected abstract List<ServletContextHandler> initialiseAndCreateContexts();

    /**
     * Returns a default handler collection that calls
     * {@link #getHandlerList(org.eclipse.jetty.server.Server, java.util.List)} to populate its
     * default handlers.
     *
     * @param server
     * @param contexts
     *
     * @return
     */
    public HandlerCollection getHandlerCollection(Server server, List<ServletContextHandler> contexts) {
        HandlerCollection handlerCollection = new HandlerCollection();
        handlerCollection.addHandler(getHandlerList(server, contexts));
        // TODO log is disabled: create external config to enable it conditionally
        // handlerCollection.addHandler(LogHandlerFactory.createLogHandler(logDir));
        return handlerCollection;
    }


    /**
     * Returns the default handler list.  These are wired in to the handler
     * collection returned by {@link #getHandlerCollection(org.eclipse.jetty.server.Server, java.util.List)}
     * so don't need to be done by the caller.
     *
     * @param server
     * @param contexts
     *
     * @return
     */
    public HandlerList getHandlerList(Server server, List<ServletContextHandler> contexts) {
        HandlerList handlers = new HandlerList();
        handlers.addHandler(new ShutdownHandler(server, secret));
        // add the dynamicContexts
        for (ServletContextHandler context : contexts) {
            handlers.addHandler(context);
        }

        return handlers;
    }

    /**
     * To be called internally by the subclasses.
     *
     * @param args
     */
    public void execute(String[] args) {
        if (args.length != 1) {
            startServer();
        } else if ("stop".equals(args[0])) {
            stopServer();
        } else if ("start".equals(args[0])) {
            startServer();
        } else {
            usage();
        }
    }

    /** Start Up the server */
    private void startServer() {
        Server srv = new Server();
        srv.setStopAtShutdown(true);
        srv.setGracefulShutdown(5000);
        QueuedThreadPool threadPool = new QueuedThreadPool();
        threadPool.setMaxThreads(workerThreadCount);
        srv.setThreadPool(threadPool);
        Connector connector = new SelectChannelConnector();
        connector.setPort(port);
        connector.setMaxIdleTime(30000);
        srv.setConnectors(new Connector[]{connector});

        try {
            List<ServletContextHandler> contexts = initialiseAndCreateContexts();
            HandlerCollection handlerCollection = getHandlerCollection(srv, contexts);
            srv.setHandler(handlerCollection);

            srv.start();
            srv.join();
        } catch (Exception e) {
            log.error("Exception: ", e);
            e.printStackTrace();
        }
    }

    /**
     *
     */
    private void stopServer() {
        HttpClient client = new HttpClient();
        try {
            PostMethod method = new PostMethod(String.format("http://127.0.0.1:%s/shutdown?secret=%s", port, secret));
            client.executeMethod(method);
            log.info(method.getResponseBodyAsString());
        } catch (IOException e) {
            log.error("Can't stop server gracefully", e);
        }
    }

    /**
     *
     */
    private void usage() {
        log.info("Usage: java -jar <file.jar> [start|stop]\n\t" + "start    Start the server (default)\n\t" + "stop     Stop the server gracefully\n");
        System.exit(-1);
    }


    /*                                           *
     *       WebServer implementation            *
     *                                           */

    @Override
    public int getPort() {
        return port;
    }

    @Override
    public void start() {
        startServer();
    }

    @Override
    public void shutdown() {
        stopServer();
    }

}
